#include "pch.h"
#include "address.h"

CAddress::CAddress()
: m_zone(-1), m_net(-1), m_node(-1), m_point(-1), m_domain()
{
}

CAddress::CAddress(int zone, int net, int node, int point, const tstring& domain)
: m_zone(zone), m_net(net), m_node(node), m_point(point), m_domain(domain)
{
}

CAddress::CAddress(const CAddress& aCopy)
: m_zone(aCopy.m_zone), m_net(aCopy.m_net), m_node(aCopy.m_node), m_point(aCopy.m_point),
  m_domain(aCopy.m_domain)
{
}

CAddress::~CAddress(void)
{
}

CAddress& CAddress::operator =(const CAddress& address) {
	m_zone   = address.m_zone;
	m_net    = address.m_net;
	m_node   = address.m_node;
	m_point  = address.m_point;
	m_domain = address.m_domain;

	return (*this);
}

LPCTSTR	FIDONET     = TEXT("fidonet");
LPCTSTR FIDONET_ORG = TEXT("fidonet.org");

bool CAddress::operator ==(const CAddress& address) const {
	if (m_zone == address.m_zone && m_net == address.m_net && m_node == address.m_node &&
	    m_point == address.m_point)
	{
		// check domain
		if (lstrcmpi( m_domain.c_str(), address.m_domain.c_str() ) == 0) {
			return true;
		} else {
			// check two combinations: fidonet & fidonet.org; fidonet.org & fidonet
			if (lstrcmpi( m_domain.c_str(), FIDONET ) == 0 && lstrcmpi( address.m_domain.c_str(), FIDONET_ORG ) == 0) {
				return true;
			}
			if (lstrcmpi( m_domain.c_str(), FIDONET_ORG ) == 0 && lstrcmpi( address.m_domain.c_str(), FIDONET ) == 0) {
				return true;
			}
			return false;
		}
	} else {
		return false;
	}
}

/**
 * Parses any addresses:
 *	node
 *	node.point
 *	.point
 *	net/node
 *	net/node.point
 *	zone:net/node
 *	zone:net/node@domain
 *	zone:net/node.point
 *	zone:net/node.point@domain
 */
bool CAddress::parseString(LPCTSTR str) {
	tstring	suspect( str );
	if (suspect.length() == 0) {
		return false;
	}

	int		zone = -1, net = -1, node = -1, point = -1;
	tstring	domain;

	int	pos, scanResult;

	if (suspect[0] == TEXT('.')) {
		// .point
		if (suspect.length() <= 1) {
			return false;
		}
		suspect = suspect.substr( 1 );
		scanResult = sscanf( suspect.c_str(), "%d", &point );
		if (scanResult == 0 || scanResult == EOF || point < 0) {
			return false;
		}

		m_zone  = -1;
		m_net   = -1;
		m_node  = -1;
		m_point = point;
		m_domain.clear();

		return true;
	}

	// 1-st symbol is not '.'
	
	// check ':'
	pos = suspect.find( TEXT(':') );
	if (pos >= 0) {
		// zone can present
		tstring	zoneSuspect = suspect.substr( 0, pos );
		scanResult = sscanf( zoneSuspect.c_str(), "%d", &zone );
		if (scanResult == 0 || scanResult == EOF || zone < 0) {
			return false;
		}
		suspect = suspect.substr( pos+1 );
	}
	// check '/'
	pos = suspect.find( TEXT('/') );
	if (pos >= 0) {
		// net awaited
		tstring	netSuspect = suspect.substr( 0, pos );
		scanResult = sscanf( netSuspect.c_str(), "%d", &net );
		if (scanResult == 0 || scanResult == EOF || net < 0) {
			return false;
		}
		suspect = suspect.substr( pos+1 );
	}
	// at this moment possible cases are:
	// node
	// node.point
	// node.point@domain

	// check '@'
	pos = suspect.find( TEXT('@') );
	if (pos >= 0) {
		// collect domain
		domain = suspect.substr( pos+1  );
		if (domain.length() == 0) {
			return false;
		}
		suspect = suspect.substr( 0, pos );
	}

	// check '.'
	pos = suspect.find( TEXT('.') );
	if (pos >= 0) {
		// collect point
		tstring	pointSuspect = suspect.substr( pos+1 );
		scanResult = sscanf( pointSuspect.c_str(), "%d", &point );
		if (scanResult == 0 || scanResult == EOF || point < 0) {
			return false;
		}
		suspect = suspect.substr( 0, pos );
	} else {
		point = 0;
	}

	// get node
	scanResult = sscanf( suspect.c_str(), "%d", &node );
	if (scanResult == 0 || scanResult == EOF || node < 0) {
		return false;
	}

	m_zone   = zone;
	m_net    = net;
	m_node   = node;
	m_point  = point;
	m_domain = domain;

	// domain can be specified only when all with/except point specified
	if (m_domain.length() > 0 && (m_zone < 0 || m_net < 0 || m_node < 0 || m_point < 0)) {
		return false;
	}

	// substitution
	if (lstrcmp( m_domain.c_str(), FIDONET ) == 0) {
		m_domain = FIDONET_ORG;
	}

	return true;
}

bool CAddress::isDefined() const {
	return (m_zone != -1 || m_net != -1 || m_node != -1 || m_point != -1 || m_domain.length() > 0);
}

tstring CAddress::toString() const {
	tstring	str;

	TCHAR	buffer[16];
	if (m_zone >= 0) {
		_stprintf( buffer, TEXT("%d:"), m_zone );
		str = buffer;
	}
	if (m_net >= 0) {
		_stprintf( buffer, TEXT("%d/"), m_net );
		str += buffer;
	}
	if (m_node >= 0) {
		_stprintf( buffer, TEXT("%d"), m_node );
		str += buffer;
	}
	// strong "above" sign for point => zero point will not be printed except for .0 address
	if (m_point > 0 || (m_point == 0 && str.length() == 0)) {
		_stprintf( buffer, TEXT(".%d"), m_point );
		str += buffer;
	}
	if (m_domain.length() > 0) {
		str += tstring( TEXT("@") ) + m_domain;
	}

	return str;
}

tstring CAddress::toStringWithoutDomain() const {
	tstring	str;

	TCHAR	buffer[16];
	if (m_zone >= 0) {
		_stprintf( buffer, TEXT("%d:"), m_zone );
		str = buffer;
	}
	if (m_net >= 0) {
		_stprintf( buffer, TEXT("%d/"), m_net );
		str += buffer;
	}
	if (m_node >= 0) {
		_stprintf( buffer, TEXT("%d"), m_node );
		str += buffer;
	}
	// strong "above" sign for point => zero point will not be printed except for .0 address
	if (m_point > 0 || (m_point == 0 && str.length() == 0)) {
		_stprintf( buffer, TEXT(".%d"), m_point );
		str += buffer;
	}

	return str;
}

// get undefined parts of the address from base address
void CAddress::baseUpon(const CAddress& baseAddress) {
	if (m_zone < 0) {
		m_zone = baseAddress.m_zone;
		if (m_net < 0) {		
			m_net = baseAddress.m_net;
			if (m_node < 0) {
				m_node = baseAddress.m_node;
			}
		}
	}
	if (m_domain.length() == 0) {
		m_domain = baseAddress.m_domain;
	}
}

tstring CAddress::toShortestString(const CAddress& baseAddr) const {
	tstring	strAddr;
	
	if (lstrcmpi( baseAddr.getDomain().c_str(), m_domain.c_str() ) != 0) {
		strAddr = toString();
	}
	else {
		TCHAR	buffer[256];

		if (m_zone == baseAddr.getZone()) {
			if (m_net == baseAddr.getNet()) {
				if (m_node == baseAddr.getNode()) {
					if (m_point == 0) {
						_stprintf( buffer, TEXT("%ld"), m_node );
					} else {
						buffer[0] = TEXT('\0');
					}
				}
				else {
					_stprintf( buffer, TEXT("%ld"), m_node );
				}
			}
			else {
				_stprintf( buffer, TEXT("%ld/%ld"), m_net, m_node );
			}
		}
		else {
			_stprintf( buffer, TEXT("%ld:%ld/%ld"), m_zone, m_net, m_node );
		}
		if (m_point != 0) {
			_stprintf( buffer + lstrlen(buffer), TEXT(".%ld"), m_point );
		}
		strAddr = buffer;
	}

	return strAddr;
}
