#include "pch.h"
#include "Misc.hpp"
#include "AreasManager.h"
#include "LinksManager.h"
#include "AreasSAX.h"

#include "Adrenalin.h"

/**********************
  CAreasContentHandler
 **********************/


CAreasContentHandler::CAreasContentHandler(CMiniSAXParser* aParser) {
	parser = aParser;
	state  = ROOT_STATE;
	group  = NULL;
	area   = NULL;
}


CAreasContentHandler::~CAreasContentHandler() {
}


static const LPCTSTR	RENAMEFILEPLAN_LEXEM = "renameFilePlan";


void CAreasContentHandler::startElement(const char* name, CAttributeList& attributes)
        throw (CMiniSAXException)
{
	State	oldState = state;
	switch (state) {
	case ROOT_STATE:
		if (strcmp( name, "areas" ) == 0) {
			state = AREAS_STATE;
			break;
		}
		break;
	case AREAS_STATE:
		if (strcmp( name, "group" ) == 0) {
			// name (required)
			const CAttribute*	nameAttr = attributes.getAttribute( "name" );
			if (nameAttr == NULL) {
				parser->raiseEx( "Group name expected (as attribute)" );
			}
			if (CAreasManager::getInstance()->getGroup( nameAttr->getValue() ) != NULL) {
				parser->raiseEx( "Found duplicate group" );
			}

			// description (optional)
			const CAttribute*	descAttr = attributes.getAttribute( "description" );
			
			group = new CAreaGroup();
			group->setName( nameAttr->getValue() );
			if (descAttr != NULL) {
				group->setDescription( descAttr->getValue() );
			}

			CAreasManager::getInstance()->addGroup( group );

			state = GROUP_STATE;
			break;
		}
	case GROUP_STATE:
		if (strcmp( name, "area" ) == 0) {
			// name (required)
			const CAttribute*	nameAttr = attributes.getAttribute( "name" );
			if (nameAttr == NULL) {
				parser->raiseEx( "Area name expected (as attribute)" );
			}
			// TODO: check name length, check name integrity (no spaces!)

			// description (optional)
			const CAttribute*	descAttr = attributes.getAttribute( "description" );
			
			// path (required)
			const CAttribute*	pathAttr = attributes.getAttribute( "path" );
			if (pathAttr == NULL) {
				parser->raiseEx( "Path expected (as attribute)" );
			}
			// send priority (optional)
			int	sendPriority = 0;
			const CAttribute*	sendPriorityAttr = attributes.getAttribute( "sendPriority" );
			if (sendPriorityAttr != NULL) {
				int	scanResult = sscanf( sendPriorityAttr->getValue(), "%d", &sendPriority );
				if (scanResult == 0 || scanResult == EOF || sendPriority < 0) {
					parser->raiseEx( "Invalid send priority" );
				}
			}
			// receive priority (optional)
			int	receivePriority = 0;
			const CAttribute*	receivePriorityAttr = attributes.getAttribute( "receivePriority" );
			if (receivePriorityAttr != NULL) {
				int	scanResult = sscanf( receivePriorityAttr->getValue(), "%d", &receivePriority );
				if (scanResult == 0 || scanResult == EOF || receivePriority < 0) {
					parser->raiseEx( "Invalid receive priority" );
				}
			}

			area = new CArea();
			// name
			area->setName( nameAttr->getValue() );
			// description
			if (descAttr != NULL) {
				area->setDescription( descAttr->getValue() );
			}
			area->setPath( pathAttr->getValue() );
			area->setSendPriority( sendPriority );
			area->setReceivePriority( receivePriority );

			// these properties can be changed by inner elements
			area->resetFlag( CArea::FLAG_PASSTHROUGH       );
			area->resetFlag( CArea::FLAG_VETO_MANUAL_PURGE );
			area->resetFlag( CArea::FLAG_OPEN_FOR_RESEND   );

			state = AREA_STATE;
			break;
		}
		break;
	case AREA_STATE:
		if (strcmp( name, "passthrough" ) == 0) {
			area->setFlag( CArea::FLAG_PASSTHROUGH );
			state = PASSTHROUGH_STATE;
			break;
		}
		if (strcmp( name, "vetoManualPurge" ) == 0) {
			area->setFlag( CArea::FLAG_VETO_MANUAL_PURGE );
			state = VETO_MANUAL_PURGE_STATE;
			break;
		}
		if (strcmp( name, "openForResend" ) == 0) {
			area->setFlag( CArea::FLAG_OPEN_FOR_RESEND );
			state = OPEN_FOR_RESEND_STATE;
			break;
		}
		if (strcmp( name, "outbound" ) == 0) {
			const CAttribute*	typeAttr = attributes.getAttribute( "type" );
			if (typeAttr != NULL) {
				if (strcmp( typeAttr->getValue(), "binkley" ) == 0) {
					area->setOutboundDefined( true );
					area->setOutboundType( OT_BINKLEY );
				} else
				if (strcmp( typeAttr->getValue(), "fileBoxes" ) == 0) {
					area->setOutboundDefined( true );
					area->setOutboundType( OT_FILE_BOXES );
				} else {
					parser->raiseEx( "Unknown outbound type" );
				}
			} else {
				parser->raiseEx( "type attribute absent" );
			}

			state = OUTBOUND_STATE;
			break;
		}
		if (strcmp( name, "fileFlag" ) == 0) {
			const CAttribute*	maskAttr = attributes.getAttribute( "mask" );
			if (maskAttr != NULL) {
				fileFlagMask = maskAttr->getValue();
				if (fileFlagMask.empty()) {
					parser->raiseEx( "Mask cannot be empty" );
				}
			} else {
				parser->raiseEx( "mask attribute absent" );
			}
			state = FILEFLAG_STATE;
			break;
		}
		if (strcmp( name, RENAMEFILEPLAN_LEXEM ) == 0) {

			const CAttribute*	typeAttr = attributes.getAttribute( "type" );
			if (typeAttr == NULL) {
				parser->raiseEx( "type attribute of renamePlan absent" );
			}

			const char*		value = typeAttr->getValue();
			RenameFilePlan	plan = RFP_ABSENT;
			if (strcmp( value, "none" ) == 0) {
				plan = RFP_NONE;
			} else
			if (strcmp( value, "increaseExt" ) == 0) {
				plan = RFP_INCREASE_EXT;
			} else
			if (strcmp( value, "addExtPrefix" ) == 0) {
				plan = RFP_ADD_EXT_PREFIX;
			} else
			if (strcmp( value, "addExtSuffix" ) == 0) {
				plan = RFP_ADD_EXT_SUFFIX;
			} else {
				parser->raiseEx( (std::string("unknown rename file plan: ") + value).c_str() );
			}

			area->setRenameFilePlan( plan );
			
			state = AREA_RENAMEFILEPLAN_STATE;
			break;
		}
		if (strcmp( name, "linked" ) == 0) {
			state = LINKED_STATE;

			// remember routing for link if specified

			const CAttribute*	routingAttr = attributes.getAttribute( "routing" );
			if (routingAttr != NULL) {
				if (strcmp( routingAttr->getValue(), "send" ) == 0) {
					linkRouting = ROUTE_SEND;
				} else
				if (strcmp( routingAttr->getValue(), "receive" ) == 0) {
					linkRouting = ROUTE_RECEIVE;
				} else
				if (strcmp( routingAttr->getValue(), "bidirectional" ) == 0) {
					linkRouting = ROUTE_BIDIRECT;
				} else {
					parser->raiseEx( "Unknown routing" );
				}				
			} else {
				// individual routing not specified
				linkRouting = ROUTE_NONE;
			}
			 
			break;
		}
		break;
	}
	if (state == oldState) {
		// if we are here then we have unexpected element
		parser->raiseEx( "Unexpected element" );
	}
	parser->checkUnexpected( attributes );
}


void CAreasContentHandler::endElement(const char* name)
        throw (CMiniSAXException)
{
	switch (state) {
	case AREAS_STATE:
		if (strcmp( name, "areas" ) == 0) {
			state = ROOT_STATE;
			return;
		}
		break;
	case GROUP_STATE:
		if (strcmp( name, "group" ) == 0) {
			group = NULL;
			state = AREAS_STATE;
			return;
		}
		break;
	case LINKED_STATE:
		if (strcmp( name, "linked" ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case AREA_RENAMEFILEPLAN_STATE:
		if (strcmp( name, RENAMEFILEPLAN_LEXEM ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case FILEFLAG_STATE:
		if (strcmp( name, "fileFlag" ) == 0) {
			if (!fileFlagMask.empty()) {
				// there was no characters in element
				// if they were we cleared the mask
				parser->raiseEx( "Flag file name wasn't specified" );
			}
			state = AREA_STATE;
			return;
		}
		break;
	case OUTBOUND_STATE:
		if (strcmp( name, "outbound" ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case OPEN_FOR_RESEND_STATE:
		if (strcmp( name, "openForResend" ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case VETO_MANUAL_PURGE_STATE:
		if (strcmp( name, "vetoManualPurge" ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case PASSTHROUGH_STATE:
		if (strcmp( name, "passthrough" ) == 0) {
			state = AREA_STATE;
			return;
		}
		break;
	case AREA_STATE:
		if (strcmp( name, "area" ) == 0) {
			area->setGroup( group );
			CAreasManager::getInstance()->addArea( area );
			area  = NULL;
			state = GROUP_STATE;
			return;
		}
		break;
	}
	// if we are here then we have unexpected element
	parser->raiseEx( "Unexpected end of element" );
}


void CAreasContentHandler::handleCharacters(const char* buffer, size_t len)
        throw (CMiniSAXException)
{
	switch (state) {
		case LINKED_STATE:
			{
				CAddress	linkAddress;
				if (!linkAddress.parseString( tstring( buffer, len).c_str() )) {
					parser->raiseEx( "Invalid link address" );
				}
				linkAddress.baseUpon( CCommonData::getInstance()->getSysopAddress() );

				PLink	link = CLinksManager::getInstance()->getLink( linkAddress );
				if (link == NULL) {
					parser->raiseEx( "Unknown link" );
				}

				TiePtr	tie = new CTie();
				tie->setLink( link );
				tie->setRouteType( linkRouting );

				area->addTie( tie );
			}
			return;
		case FILEFLAG_STATE:
			{				
				if (!hasOnlySpaces( buffer, len ) && !fileFlagMask.empty()) {
					area->addFileFlag( CMaskFileFlag( fileFlagMask, tstring( buffer, len ) ) );
					fileFlagMask.clear();
				}
			}
			return;
	}
	if (!hasOnlySpaces( buffer, len )) {
		parser->raiseEx( "Unexpected symbols" );
	}
}
