#include "pch.h"
#include "logger.h"

ModeHolder	LOG_START(   LM_START   );
ModeHolder	LOG_STOP(    LM_STOP    );
ModeHolder	LOG_SILENT(  LM_SILENT  );
ModeHolder	LOG_LOUD(    LM_LOUD    );
ModeHolder	LOG_WARNING( LM_WARNING );
ModeHolder	LOG_ERROR(   LM_ERROR   );

CLogger::CLogger(void) : m_mode(LM_LOUD), m_startOfLine(true)
{
}

CLogger::~CLogger(void)
{
}

bool CLogger::openLogFile(LPCTSTR logFileName) {
	m_logFile.open( logFileName, std::ios::app );
	if (m_logFile.fail()) {
		m_logFile.open( logFileName );
	}

	return (!m_logFile.fail());
}

void CLogger::logToStdOut(LPCTSTR message) {
	std::cout << message;
}

void CLogger::logToStdErr(LPCTSTR message) {
	std::cerr << message;
}

void CLogger::logToFile(LPCTSTR message) {
	m_logFile << message;
}

CLogger& CLogger::operator << (const tstring& message) {
	return ((*this) << message.c_str());
}

CLogger& CLogger::operator << (const CAddress& address) {
	return ((*this) << address.toString());
}

CLogger& CLogger::operator << (LPCTSTR message) {
	tstring	part;

	int	fromPos = 0;
	int	toPos   = fromPos;
	while (message[toPos] != TEXT('\0')) {
		if (message[toPos] == TEXT('\n')) {
			// print message part [fromPos,toPos]
			part.assign( message + fromPos, toPos - fromPos + 1 );
			logSingleLinePart( part.c_str() );
			m_startOfLine = true;
			fromPos = toPos + 1;
		}
		++toPos;
	}
	if (fromPos < toPos) {
		part.assign( message + fromPos, toPos - fromPos );
		logSingleLinePart( part.c_str() );
	}

	return (*this);
}

CLogger& CLogger::operator << (int value) {
	TCHAR	buffer[128];
	_stprintf( buffer, TEXT("%ld"), value );
	return ((*this) << buffer);
}

CLogger& CLogger::operator << (unsigned short value) {
	TCHAR	buffer[128];
	_stprintf( buffer, TEXT("%hu"), value );
	return ((*this) << buffer);
}

void CLogger::flush() {
	std::cout.flush();
	std::cerr.flush();
	m_logFile.flush();
}

LPCTSTR START_PREFIX   = TEXT("[  START] ");
LPCTSTR STOP_PREFIX    = TEXT("[   STOP] ");
LPCTSTR INFO_PREFIX    = TEXT("[   info] ");
LPCTSTR	WARNING_PREFIX = TEXT("[warning] ");
LPCTSTR	ERROR_PREFIX   = TEXT("[  error] ");

void CLogger::logSingleLinePart(LPCTSTR part) {
	if (m_startOfLine) {
		switch (m_mode) {
			case LM_START:
				logToFile( START_PREFIX );
				break;
			case LM_STOP:
				logToFile( STOP_PREFIX );
				break;
			case LM_SILENT:
				logToFile( INFO_PREFIX );
				break;
			case LM_LOUD:
				logToStdOut( INFO_PREFIX );
				logToFile( INFO_PREFIX );
				break;
			case LM_WARNING:
				logToStdOut( WARNING_PREFIX );
				logToFile(   WARNING_PREFIX );
				break;
			case LM_ERROR:
				logToStdErr( ERROR_PREFIX );
				logToFile(   ERROR_PREFIX );
				break;
		}
		m_startOfLine = false;
	}
	switch (m_mode) {
		case LM_START:
		case LM_STOP:
		case LM_SILENT:
			logToFile( part );
			break;
		case LM_LOUD:
		case LM_WARNING:
			logToStdOut( part );
			logToFile( part );
			break;
		case LM_ERROR:
			logToStdErr( part );
			logToFile( part );
			break;
	}
}

CLogger& CLogger::operator << (const ModeHolder modeHolder) {
	// if new mode and not start of line then we should switch to new line
	if (modeHolder.getMode() != m_mode && !m_startOfLine) {
		logSingleLinePart( TEXT("\n") );
		m_startOfLine = true;
	}
	m_mode = modeHolder.getMode();

	return (*this);
}

void CLogger::rememberMode() {
	m_modeHistory.push( m_mode );
}

void CLogger::restoreMode() {
	if (!m_modeHistory.empty()) {
		LogMode	mode = m_modeHistory.top();
		m_modeHistory.pop();
		// if new mode and not start of line then we should switch to new line
		if (m_mode != mode && !m_startOfLine) {
			logSingleLinePart( TEXT("\n") );
			m_startOfLine = true;
		}
		m_mode = mode;
	}
}


void CLogger::formatDateTime(const SYSTEMTIME& someTime, LPTSTR timeString) {
	_stprintf( timeString, TEXT("%04hu/%02hu/%02hu %02hu:%02hu:%02hu"),
			   someTime.wYear, someTime.wMonth,  someTime.wDay,
			   someTime.wHour, someTime.wMinute, someTime.wSecond );
}


void CLogger::destroy() {
	// log stop time
	SYSTEMTIME	stopTime;
	::GetLocalTime( &stopTime );
	TCHAR	buffer[128];
	CLogger::formatDateTime( stopTime, buffer );
	(*this) << LOG_STOP << buffer << TEXT("\n");

	// close log file
	m_logFile.close();
}
