// MiniSAX.cpp : Defines the entry point for the DLL application.
//

#include "stdafx.h"
#include "MiniSAX.h"

#include "MiniSAXImpl.h"
#include "AttributeImpl.h"

#define DEFINED(a) ((a) != NULL)


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ulReasonForCall, 
                       LPVOID lpReserved
					 )
{
    switch (ulReasonForCall)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }
    return TRUE;
}


using namespace miniSAX;


/*****************************
  CIOException implementation
 *****************************/


CIOException::CIOException(const char* aMessage) {
	if (DEFINED(aMessage)) {
		message = new char[strlen(aMessage) + 1];
		strcpy( message, aMessage );
	}
	else {
		message = NULL;
	}
}


CIOException::~CIOException() {
	if (DEFINED(message)) {
		delete [] message;
		message = NULL;
	}
}


/****************************
  CDataSource implementation
 ****************************/


CDataSource::CDataSource(const char* fileName) throw (CIOException) {
	loadFromFile( fileName );
}


CDataSource::~CDataSource() {
	freeData();
}


void CDataSource::freeData() {
	if (DEFINED(data)) {
		free(data);
		len    = 0;
		offset = 0;
		col    = 1;
		row    = 1;
	}
}


void CDataSource::loadFromFile(const char* fileName)
	throw (CIOException)
{
	const size_t	BUFFER_SIZE = 32*1024;

	// set startup data buffer
	size_t	bufferSize = BUFFER_SIZE;
	data   = (char*)malloc( bufferSize );
	len    = 0;
	offset = 0;
	col    = 1;
	row    = 1;

	// open file
	HANDLE	fileHandle = ::CreateFile( fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );
	if (fileHandle == INVALID_HANDLE_VALUE) {
		throw CIOException( "Failed to open file for reading" );
	}

	// prepare read buffer
	char*	buffer = (char*)malloc( BUFFER_SIZE );

	// read file
	DWORD	bytesRead;
	do {
		if (!::ReadFile( fileHandle, buffer, BUFFER_SIZE, &bytesRead, NULL )) {
			break;
		}
		if (bytesRead > 0) {
			// check data size that can be used
			if ((bufferSize - len) < bytesRead) {
				size_t	needed = len + bytesRead;
				bufferSize <<= 1;
				while (bufferSize < needed) {
					bufferSize <<= 1;
				}
				char*	newData = (char*)realloc( data, bufferSize );
				if (!DEFINED(newData)) {
					// TODO: handle error
					::CloseHandle( fileHandle );
					return;
				}
				data = newData;
			}
			memcpy( data + len, buffer, bytesRead );
			len += bytesRead;
		}
	} while (bytesRead > 0);

	// free read buffer
	free( buffer );

	// close file
	::CloseHandle( fileHandle );
}


bool CDataSource::hasData() const {
	return (offset < len);
}


char CDataSource::getChar() {
	// TODO: check offset (data presence)
	char	currentChar = data[offset++];
	if (currentChar == '\n') {
		col++;
		row=1;
	}
	else {
		row++;
	}
	return currentChar;
}


void CDataSource::skipSpaces() {
	while (offset < len) {
		char	currentChar = data[offset];
		if (currentChar != ' ' && currentChar != '\n' && currentChar != '\r' && currentChar != '\t') {
			break;
		}
		if (currentChar == '\n') {
			col++;
			row=1;
		}
		else {
			row++;
		}
		offset++;
	}
}


size_t	CDataSource::getDistance(char delimiter) const {
	int	distance = offset;
	while (distance < len && data[distance] != delimiter) {
		distance++;
	}

	return (distance - offset);
}


size_t CDataSource::getNonElementDistance() const {
	int	distance = offset;
	while (distance < len) {
		char	currentChar = data[distance];
		if (isalpha(currentChar) || isdigit(currentChar) || currentChar == '_' || currentChar == '-') {
			distance++;
		}
		else {
			break;
		}
	}

	return (distance - offset);
}


void CDataSource::getSequence(char* dest, int seqLen) {
	// TODO: check boundaries
	if (seqLen > 0) {

		// move position
		int	from = offset;
		int	to   = offset + seqLen - 1;
		for (int i = from; i <= to; i++) {
			if (data[i] == '\n') {
				col++;
				row=1;
			}
			else {
				row++;
			}
		}

		memcpy( dest, data + offset, seqLen );
		offset += seqLen;
	}
}


void CDataSource::back() {
	offset--;
	if (data[offset] == '\n') {
		col--;
		// restore row value
		int	rowOffset = offset-1;
		while (rowOffset > 0 && data[rowOffset] != '\n') {
			--rowOffset;
		}
		row = offset - rowOffset;
	}
	else {
		row--;
	}
}


/*******************
  CMiniSAXException
 *******************/


CMiniSAXException::CMiniSAXException(const char* aMessage, int aCol, int aRow) {
	message = new char[strlen(aMessage)+1];
	strcpy( message, aMessage );
	col = aCol; row = aRow;
}


CMiniSAXException::CMiniSAXException(const char* aMessage) {
	message = new char[strlen(aMessage)+1];
	strcpy( message, aMessage );
	col = -1; row = -1;
}


CMiniSAXException::CMiniSAXException(const CMiniSAXException& ex) {
	message = new char[strlen(ex.message)+1];
	strcpy( message, ex.message );
	col = ex.col; row = ex.row;
}


CMiniSAXException::~CMiniSAXException() {
	delete [] message;
}


/****************
  CAttributeList
 ****************/


CAttributeList::CAttributeList() {
}


CAttributeList::~CAttributeList() {
}


/*******************************
  CMiniSAXParser implementation
 *******************************/

CMiniSAXParser::CMiniSAXParser() {
}


CMiniSAXParser::~CMiniSAXParser() {
}


CMiniSAXParser* CMiniSAXParser::create() {
	return CMiniSAXParserImpl::create();
}
