#pragma	once

#ifdef MINISAX_EXPORTS
#define MINISAX_API __declspec(dllexport)
#else
#define MINISAX_API __declspec(dllimport)
#endif

namespace miniSAX {


	class MINISAX_API CIOException {
	public:
		CIOException(const char* aMessage);
		virtual ~CIOException();

		const char* getMessage() const { return message; }
	protected:
		char*	message;
	};


	class MINISAX_API CDataSource {
	public:
		CDataSource(const char* fileName) throw (CIOException);
		virtual ~CDataSource();

		virtual bool hasData() const;
		virtual char getChar();
		virtual void skipSpaces();

		virtual size_t	getDistance(char delimiter) const;
		virtual size_t	getNonElementDistance() const;

		virtual void getSequence(char* dest, int seqLen);

		virtual void back();

		virtual int getColumn() const { return col; }
		virtual int getRow() const { return row; }

	protected:
		char*	data;
		int		len;
		int		offset;

		int		col, row;

		virtual void freeData();
		virtual void loadFromFile(const char* fileName) throw (CIOException);
	};


	class MINISAX_API CMiniSAXException {
	public:
		/**
		 * Exception at the known position
		 */
		CMiniSAXException(const char* aMessage, int aCol, int aRow);
		/**
		 * Exception at an unknown position
		 */
		CMiniSAXException(const char* aMessage);
		CMiniSAXException(const CMiniSAXException& ex);
		virtual ~CMiniSAXException();

		const char* getMessage() const { return message; }
		int			getColumn()  const { return col;     }
		int			getRow()     const { return row;     }

	protected:
		char*	message;
		int		col;
		int		row;
	};


	class MINISAX_API CAttribute {
	public:
		CAttribute() {}
		virtual ~CAttribute() {}

		virtual const char* getName()  const = 0;
		virtual const char* getValue() const = 0;
	};


	class MINISAX_API CAttributeList {
	public:
		CAttributeList();
		virtual ~CAttributeList();

		virtual void addAttribute(const CAttribute& newAttr) = 0;

		virtual size_t getCount() const = 0;
		virtual const CAttribute* getAttribute(int index) const = 0;

		virtual const CAttribute* getAttribute(const char* name) = 0;
		// following method allows to get n-th attribute with the same name
		virtual const CAttribute* getAttribute(const char* name, size_t n) = 0;
		// Each request of attribute via getAttribute(name) prepares list of requested attributes
		// When we finish with known attributes we can call checkUnexpected thus
		// allowing to throw exception if there still some unexpected attributes
		// NOTE: this method can be called only once after processing because it will delete list of requests
		// TODO: allow to request many attributes with the same name
		virtual void checkUnexpected() throw(CMiniSAXException) = 0;
	};


	class MINISAX_API CDocumentHandler {
	public:
		virtual void startDocument() throw (CMiniSAXException) = 0;
		virtual void endDocument() throw (CMiniSAXException) = 0;
	};


	class MINISAX_API CContentHandler {
	public:
		virtual void startElement(const char* name, CAttributeList& attributes)
			 throw (CMiniSAXException) = 0;
		virtual void endElement(const char* name)
			 throw (CMiniSAXException) = 0;

		virtual void handleCharacters(const char* buffer, size_t len)
			 throw (CMiniSAXException) = 0;
	};


	class MINISAX_API CErrorHandler {
	public:
		virtual void handleError(const CMiniSAXException& e) = 0;
	};


	class MINISAX_API CEntityResolver {
	public:
		virtual void resolveEntity()
			throw (CMiniSAXException) = 0;
	};


	// this class can be created only using class factory
	// class factory is integrated in it thus allowing to just call CMiniSAXParser::create
	class MINISAX_API CMiniSAXParser {
	public:
		virtual ~CMiniSAXParser();

		// Sets encoding that will be used to convert data to before submitting
		// any data to handlers.
		// Currently supported encoding is "cp866".
		// Exception is thrown when caller tries to set unsupported encoding
		virtual void setHandlerEncoding(const char* encoding) throw(CMiniSAXException) = 0;
		// Gets original encoding of the document.
		// Returns NULL if encoding was not specified.
		virtual const char* getDocumentEncoding() const = 0;

		virtual void setDocumentHandler(CDocumentHandler* aHandler) = 0;
		virtual void setContentHandler(CContentHandler* aHandler) = 0;
		virtual void setErrorHandler(CErrorHandler* aHandler) = 0;
		virtual void setEntityResolver(CEntityResolver* aResolver) = 0;

		virtual void parse(CDataSource* dataSource) throw(CMiniSAXException) = 0;

		virtual int getCurrentRow()    const = 0;
		virtual int getCurrentColumn() const = 0;

		virtual void raiseEx(const char* msg) const throw(CMiniSAXException) = 0;

		// Checks unexpected attributes the same way as list checks but exception will also contain
		// current element' coordinates
		virtual void checkUnexpected(CAttributeList& list) const throw(CMiniSAXException) = 0;

		static CMiniSAXParser* create();
	protected:
		CMiniSAXParser();
	};

	#define	ENCODING_CP866			"cp866"
	#define	ENCODING_WINDOWS_1251	"Windows-1251"
	#define	ENCODING_KOI8_R			"koi8-r"

	class MINISAX_API CCharSetConvertor
	{
	public:
		CCharSetConvertor(void);
		virtual ~CCharSetConvertor(void);

		virtual void setTransformation(const char* sourceEncoding, const char* destEncoding)
			throw (CMiniSAXException);
		virtual void convert(const char* source, char* dest);
		virtual void convert(const char* source, char* dest, int len);
	protected:
		int*	m_table;
		bool	m_disposeTable;
		virtual int* getBackwardTable(int* table);
	};
}
