#include "pch.h"

#include "Mem.hpp"
#include "Common.h"
#include "CfgParse.hpp"
#include "Misc.hpp"
#include "Links.h"
#include "Areas.h"
#include "Msg.hpp"
#include "Tic.h"
#include "crc32.hpp"
#include "Temp.hpp"
#include "File.h"

#include "CfgLexems.hpp"
#include "WELexems.h"

#include "MsgAttrs.h"

#include "LinksManager.h"
#include "AreasManager.h"

#include "Logger.h"

#include "SuspendedSAX.h"

#include "Adrenalin.h"


void CAdrenalin::hatchSuspendedFiles() {
	SuspendedFiles	suspendedFiles;

	tstring	fileName = SUSPENDED_FILES_FILE_NAME;
	CAdrenalin::adjustFileName( fileName );

	if (SuspendedHandler::doesFileExist( fileName )) {
		tstring			documentEncoding;
		if (SuspendedHandler::readSuspendedFiles( fileName, &suspendedFiles, documentEncoding )) {
			// get current datetime
			SYSTEMTIME	dateTime;
			::GetLocalTime( &dateTime );
			FILETIME	fileTime, suspendedFileTime;
			::SystemTimeToFileTime( &dateTime, &fileTime );
			LARGE_INTEGER	currentDateTime, suspendedDateTime;	// will be used to compare times
			currentDateTime.LowPart  = fileTime.dwLowDateTime;
			currentDateTime.HighPart = fileTime.dwHighDateTime;
			// remember number of suspended files
			size_t	originalCount = suspendedFiles.size();

			bool	dataChanged = false;

			size_t	count = originalCount;
			size_t	i     = 0;
			while (i < count) {
				SuspendedFilePtr	file = suspendedFiles[i];
				if (!file->hasError()) {
					// process only good files
					::SystemTimeToFileTime( &(file->getReleaseDateTime()), &suspendedFileTime );
					suspendedDateTime.LowPart  = suspendedFileTime.dwLowDateTime;
					suspendedDateTime.HighPart = suspendedFileTime.dwHighDateTime;
					// compare times
					if (currentDateTime.QuadPart > suspendedDateTime.QuadPart) {
						// time to release this great file to the World!
						LoggerPtr	logger;
						logger << LOG_LOUD << TEXT("Time to release file!\n")
							<< TEXT("  file: ") << file->getFileName()
							<< TEXT("\n  area: ") << file->getAreaName() << TEXT("\n");
						int result = hatchSingleFile( file->getAreaName(), file->getFileName(),
													file->getDescription(), file->isNoCopy(),
													file->getReplaces() );
						switch (result) {
							case ERROR_NONE:
								if (file->doDeleteAfterHatch()) {
									if (_tunlink( file->getFileName().c_str() ) != 0) {
										logger << LOG_ERROR << TEXT("Failed to delete file: ")
											   << file->getFileName() << TEXT("\n");
									}
								}
								// this case allow us to remove file from the suspended files list
								suspendedFiles.erase( suspendedFiles.begin() + i );
								delete file;
								--count;
								break;
							case ERROR_AREA_NOT_FOUND:
								file->setError( TEXT("area not found") );
								++i;
								break;
							case ERROR_FILE_ABSENT:
								file->setError( TEXT("file not found") );
								++i;
								break;
							default:
								file->setError( TEXT("unknown error, hatch failed") );
								++i;
								break;
						}
						dataChanged = true;
					} else {
						// switch to next file
						++i;
					}
				} else {
					// switch to next file
					++i;
				}
			}
			if (dataChanged) {
				if (count == 0) {
					// remove suspended files file because there is no more suspended files
					_tunlink( fileName.c_str() );
				} else {
					// overwrite suspended files file
					SuspendedHandler::writeSuspendedFiles( fileName, &suspendedFiles, documentEncoding );
				}
			}
			// delete files
			while (!suspendedFiles.empty()) {
				SuspendedFilePtr	file = suspendedFiles[suspendedFiles.size()-1];
				suspendedFiles.pop_back();
				delete file;
			}
		}
	}
}


/**
 * Implementation of CAdrenalin::ProcessInboundTics()
 * @param	ticsPath	path to look for tics
 * @param	filesPath	path to look for files for tics
 */
void CAdrenalin::ProcessInboundTics(LPCTSTR ticsPath, LPCTSTR filesPath) {
	// first of all lets check temp BSO: it can contain previous failed tosses
	checkTempBSO();

	// then hatch suspended files
	hatchSuspendedFiles();

    // if path is specified for processed TICs we need to calculate their names
    CTicNamer   processedTicsNamer;

	WIN32_FIND_DATA FindFileData;
	HANDLE          hFindFile;
	BOOL            fFileFound;
	TCHAR           szFindFileMask[MAX_PATH];
	TCHAR           szFullFileName[MAX_PATH];
	TCHAR			szDestFilePath[MAX_PATH];
	TCHAR           szDestFileName[MAX_PATH];
	TCHAR           szNewAreasFileName[MAX_PATH];
	int             nProcessedTicNumber;

	VTString	vsTics;

	CTic            tic;

	vector<PLink>		forwards;

	CFile		newFilesFile;
	CFile		newAreasFile;

	UINT            uiRealCRC32;

	int             nTicNameFrom;

	unsigned int	index;

	TCHAR	szLine[LINE_SIZE];
	TCHAR   szLogLine[LINE_SIZE];
	TCHAR   szTimeLine[LINE_SIZE];

	int     *AKASeen;

	bool    fReplaceDone;
	TCHAR   szReplaceLine[LINE_SIZE];

	ifstream        in;

	CCfgParser     *TrafficParser;

	int             nTraffic;
	bool            fNewFilesArrived;
	bool            fTossedForLinks; // if yes, we should create TossFlag

	vector<tstring>	bodyLines;	// body lines
	vector<tstring>	warnings;	// warning lines
	vector<tstring>	errors;		// error lines
	CMsg		warnMessage;	// warn message to sysop

	CommonDataPtr	commonData = CCommonData::getInstance();

	// prepare log
	LoggerPtr	logger;

	if (commonData->mustShowGoOn()) {
		logger << LOG_LOUD;
    } else {
		logger << LOG_SILENT;
    }

	// print header (tossing information)

	logger << TEXT("Tossing TICs in directory ") << ticsPath << TEXT(":\n");
	logger.flush();

    // make all fecho directories and set areas traffic to zero
	{
		int	areasCount = CAreasManager::getInstance()->getAreasCount();
		for (int i = 0; i < areasCount; i++) {
			PArea	area = CAreasManager::getInstance()->getArea( i );
			MakeDirStructure( area->getWorkingPath().c_str() );
			area->setTraffic( 0 );
		}
	}

    // make tics directories
	MakeDirStructure( commonData->getBadTicsPath().c_str() );
	if (!commonData->getNewTicsPath().empty()) {
		MakeDirStructure( commonData->getNewTicsPath().c_str() );
	}
	if (!commonData->getProcessedTicsPath().empty()) {
        MakeDirStructure( commonData->getProcessedTicsPath().c_str() );
        processedTicsNamer.readTICs( commonData->getProcessedTicsPath().c_str() );
    }

    // read areas traffic
    
    lstrcpy( szFullFileName, m_szStartupPath );
    lstrcat( szFullFileName, TEXT("traffic.dat") );

    *szTimeLine = TEXT('\0');
    TrafficParser = new CCfgParser( szFullFileName );
    if (!TrafficParser->Failed()) {
        // read date line
        TrafficParser->ReadLine();
        if (!TrafficParser->Failed()) {
            lstrcpy( szTimeLine, TrafficParser->String() );
        }
        // read data lines
        while (!TrafficParser->Eof()) {
            TrafficParser->ReadLine();
            if (TrafficParser->Failed())
                break;
            if (TrafficParser->Count() != 2)
                break;
			PArea	area = CAreasManager::getInstance()->getArea( (*TrafficParser)[0] );
            if (area == NULL)
                continue;
	        if (!StrToInt( (*TrafficParser)[1], nTraffic ))
		        break;
            area->setTraffic( nTraffic );
        }
    }
    delete TrafficParser;
    
    lstrcpy( szFindFileMask, ticsPath );
    lstrcat( szFindFileMask, TEXT("*.ti?") );

    nProcessedTicNumber = 0;

    hFindFile = FindFirstFile( szFindFileMask, &FindFileData );
    fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
    while (fFileFound) {
        if (!(FindFileData.dwFileAttributes &
              (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
               FILE_ATTRIBUTE_OFFLINE | FILE_ATTRIBUTE_READONLY |
               FILE_ATTRIBUTE_TEMPORARY)))
        {
		    vsTics.push_back( ticsPath );
		    vsTics.back() += FindFileData.cFileName;
        }
        fFileFound = FindNextFile( hFindFile, &FindFileData );
    }
    if (hFindFile != INVALID_HANDLE_VALUE)
        FindClose( hFindFile );

    // allocate memory for seen AKA
	AKASeen = new int[commonData->getSysopAKACount() + 1];
    // open newfiles list
    ConcatStrings( szFullFileName, m_szStartupPath, TEXT("newfiles.txt") );
    newFilesFile.Open( szFullFileName, GENERIC_WRITE, OPEN_ALWAYS );
    newFilesFile.SeekToEnd();
    // prepare file name for new areas
    ConcatStrings( szNewAreasFileName, m_szStartupPath, TEXT("newareas.txt") );
    // do not update traffic.dat if no new files arrived
    fNewFilesArrived = false;
    fTossedForLinks  = false;
    // process TICs
    vector<tstring>::iterator	itTic = vsTics.begin();
    while (itTic != vsTics.end()) {
	    // get name of tic without path
	    {
		    LPCTSTR	szTicName = (*itTic).c_str();
			int	i;
		    for ( nTicNameFrom = -1, i = 0; szTicName[i] != _T('\0'); i++) {
		        if (szTicName[i] == _T('\\'))
			        nTicNameFrom = i;
		    }
		    nTicNameFrom++;
	    }
        // trying to parse TIC
        if (tic.Read( (*itTic).c_str() )) {
            do {
                // check link

                PLink	link = CLinksManager::getInstance()->getLink( tic.getFromAddress() );
                if (link == NULL) {
			        lstrcpy( szLine, TEXT("    unknown sender (") );
			        lstrcpy( szLine + lstrlen(szLine), tic.getFromAddress().toString().c_str() );
			        lstrcat( szLine, TEXT(")") );
			        // warn sysop if have to
					if (CCommonData::getInstance()->doNotifyOnErrors()) {
			            errors.push_back( szLine );
			        }
			        _stprintf( szLogLine, TEXT("! %12s: %s"),
				               tic.getName(), szLine + 4 );
			        logger << szLogLine << TEXT("\n");
					if (!tossingBadTics)
						MoveTicToBads( tic.getFullName(), nTicNameFrom );
			        break;
                }

                // check link activity  
                if (link->getActivity() == ACTIVITY_UNAVAILABLE) {
			        lstrcpy( szLine, TEXT("    unavailable sender (") );
			        lstrcpy( szLine + lstrlen(szLine), tic.getFromAddress().toString().c_str() );
			        lstrcat( szLine, TEXT(")") );
			        // warn sysop if have to
			        if (CCommonData::getInstance()->doNotifyOnErrors()) {
			            errors.push_back( szLine );
			        }
			        _stprintf( szLogLine, TEXT("! %12s: %s"),
				               tic.getName(),
				               szLine + 4 );
			        logger << szLogLine << TEXT("\n");
					if (!tossingBadTics)
						MoveTicToBads( tic.getFullName(), nTicNameFrom );
			        break;
                }

                // check password
                if (lstrcmpi( link->getPassword().c_str(), tic.m_szPW ) != 0) {
			        _stprintf( szLogLine,
			                   TEXT("! %12s: password error (%s instead of %s)"),
			                   tic.getName(), tic.m_szPW,
			                   link->getPassword().c_str() );
			        // warn sysop if have to
			        if (CCommonData::getInstance()->doNotifyOnErrors()) {
			            errors.push_back( szLogLine );
			        }
			        logger << szLogLine << TEXT("\n");
					if (!tossingBadTics)
						MoveTicToBads( tic.getFullName(), nTicNameFrom );
			        break;
                }

                // form full file name of inbound file
                lstrcpy( szFullFileName, filesPath );
                lstrcat( szFullFileName, tic.getFileName() );

                // check file
                hFindFile = FindFirstFile( szFullFileName, &FindFileData );
                fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
                if (!fFileFound) {
			        _stprintf( szLine, TEXT("    file %s not found"),
				               tic.getFileName() );
			        // warn sysop if have to
			        if (CCommonData::getInstance()->doNotifyOnWarnings()) {
			            warnings.push_back( szLine );
			        }
		            _stprintf( szLogLine, TEXT("! %12s: %s"),
                               tic.getName(), szLine + 4 );
			        logger << szLogLine << TEXT("\n");
					if ((CCommonData::getInstance()->getMaxTicAge() > 0) && (!tossingBadTics)) {
						if (isTicOld( tic.getFullName(), CCommonData::getInstance()->getMaxTicAge() )) {
							MoveTicToBads( tic.getFullName(), nTicNameFrom );
							lstrcpy( szLine, TEXT("    TIC was too old and moved into bad tics") );
							// warn sysop if have to
							if (CCommonData::getInstance()->doNotifyOnWarnings()) {
								warnings.push_back( szLine );
							}
							_stprintf( szLogLine, TEXT("! %12s: %s"),
									   tic.getName(), szLine + 4 );
							logger << szLogLine << TEXT("\n");
						}
					}
			        break;
                }
                FindClose( hFindFile );

                // check file finished or not
                if (FindFileData.dwFileAttributes &
                    (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
                     FILE_ATTRIBUTE_OFFLINE   | FILE_ATTRIBUTE_READONLY |
                     FILE_ATTRIBUTE_TEMPORARY))
                {
		            _stprintf( szLine, TEXT("    file %s is not accessible"),
	                           tic.getFileName() );
			        // warn sysop if have to
			        if (CCommonData::getInstance()->doNotifyOnWarnings()) {
			            warnings.push_back( szLine );
			        }
		            _stprintf( szLogLine, TEXT("! %12s: %s"),
	                           tic.getName(), szLine + 4 );
			        logger << szLogLine << TEXT("\n");
			        break;
                }

                // check size
                int	nFileSize = ((FindFileData.nFileSizeHigh * MAXDWORD) +
				                 FindFileData.nFileSizeLow);

				// remember size
				filesize_t	fileSize;
				fileSize.LowPart  = FindFileData.nFileSizeLow;
				fileSize.HighPart = FindFileData.nFileSizeHigh;

		        if (tic.getSize() != -1) {
			        if (tic.getSize() != nFileSize) {
				        _stprintf( szLine,
				                   TEXT("    file size error (%ld instead of %ld)"),				     
				                   (FindFileData.nFileSizeHigh * MAXDWORD) +
				                   FindFileData.nFileSizeLow, tic.getSize() );
				        // warn sysop if have to
				        if (CCommonData::getInstance()->doNotifyOnErrors()) {
				            errors.push_back( szLine );
				        }
				        _stprintf( szLogLine,
				             TEXT("! %12s: %s"),
				             tic.getName(),
				             szLine + 4 );
				        logger << szLogLine << TEXT("\n");
				        break;
			        }
                } else {
			        tic.setSize( nFileSize );
                }

                // check crc                
                if (!CalcCRC32( szFullFileName, uiRealCRC32 )) {
	                _stprintf( szLine,
                               TEXT("    failed to calculate CRC32 of file %s"),
                               tic.getFileName() );
			        // warn sysop if have to
			        if (CCommonData::getInstance()->doNotifyOnErrors()) {
			            errors.push_back( szLine );
			        }
	                _stprintf( szLogLine,
                               TEXT("! %12s: %s"),
                               tic.getName(), szLine + 4 );
		            logger << szLogLine << TEXT("\n");
			        break;
                }

				if (CCommonData::getInstance()->doWarnOnIncorrectCRC32()) {
			        if (tic.isCRC32Present()) {
				        if (uiRealCRC32 != tic.m_uiCRC32) {
					        _stprintf( szLine,
				                       _T("    CRC32 warning (%08lX instead of %08lX)"),
					                   uiRealCRC32,
					                   tic.getCRC32() );
					        // warn sysop if have to
					        if (CCommonData::getInstance()->doNotifyOnWarnings()) {
						        warnings.push_back( szLine );
					        }
					        _stprintf( szLogLine,
					                   _T("! %12s: %s"),
					                   tic.getName(), szLine + 4 );
							logger.rememberMode();
							logger << LOG_WARNING << szLogLine << TEXT("\n");
							logger.restoreMode();
					        // set correct CRC32
					        tic.setCRC32( uiRealCRC32 );
				        }
                    } else {
				        // tic hasn't crc32
				        lstrcpy( szLine, TEXT("    CRC32 warning (tag absent)") );
				        // warn sysop if have to
				        if (CCommonData::getInstance()->doNotifyOnWarnings()) {
					        warnings.push_back( szLine );
				        }
				        _stprintf( szLogLine, TEXT("! %12s: %s"),
					               tic.getName(), szLine + 4 );
						logger.rememberMode();
				        logger << LOG_WARNING << szLogLine << TEXT("\n");
				        // restore log output
						logger.restoreMode();
				        // set correct CRC32
				        tic.setCRC32( uiRealCRC32 );
				        tic.setCRC32Present( true );
			        }
                } else {
			        if (!tic.isCRC32Present()) {
				        lstrcpy( szLine, TEXT("    CRC32 error (tag absent)") );
				        // warn sysop if have to
				        if (CCommonData::getInstance()->doNotifyOnErrors()) {
					        errors.push_back( szLine );
				        }
				        _stprintf( szLogLine, TEXT("! %12s: %s"),
					               tic.getName(), szLine + 4 );
						logger.rememberMode();
				        logger << LOG_ERROR << szLogLine << TEXT("\n");
						logger.restoreMode();
				        break;
			        }
			        if (uiRealCRC32 != tic.getCRC32()) {
				        _stprintf( szLine,
				                   TEXT("    CRC32 error (%08lX instead of %08lX)"),
				                   uiRealCRC32, tic.getCRC32() );
				        // warn sysop if have to
				        if (CCommonData::getInstance()->doNotifyOnErrors()) {
					        errors.push_back( szLine );
				        }
				        _stprintf( szLogLine, TEXT("! %12s: %s"),
				                   tic.getName(), szLine + 4 );
						logger.rememberMode();
				        logger << LOG_ERROR << szLogLine << TEXT("\n");
						logger.restoreMode();
				        break;
			        }
		        }

		        // TODO: embed "warn sysop" code in code below

                // check area
				PArea	area = CAreasManager::getInstance()->getArea( tic.getAreaName() );
                if (area != NULL) {
					ConstTiePtr	tie = area->getTie( link );
                    if (tie == NULL) {
                        _stprintf( szLogLine, _T("! %12s: "),
				                   tic.getName() );
			            lstrcpy( szLogLine + lstrlen(szLogLine), tic.getFromAddress().toString().c_str() );
			            _stprintf( szLogLine+lstrlen(szLogLine), _T(" isn't linked to %s\r\n"),
				                   tic.getAreaName() );
						logger.rememberMode();
				        logger << LOG_ERROR << szLogLine << TEXT("\n");
						logger.restoreMode();
						if (!tossingBadTics)
							MoveTicToBads( tic.getFullName(), nTicNameFrom );
                        break;
                    }
                    if (!area->canSend( link )) {
                        _stprintf( szLogLine, _T("! %12s: "), tic.getName() );
			            lstrcpy( szLogLine + lstrlen(szLogLine), tic.getFromAddress().toString().c_str() );
			            _stprintf( szLogLine+lstrlen(szLogLine), _T(" can't send into %s\r\n"),
                                   area->getName().c_str() );
						logger.rememberMode();
				        logger << LOG_ERROR << szLogLine << TEXT("\n");
						logger.restoreMode();
						if (!tossingBadTics)
							MoveTicToBads( tic.getFullName(), nTicNameFrom );
                        break;
                    }
                } else {
                    if (link->canCreateAreas()) {
                        if (!CArea::isValidName( tic.getAreaName() )) {
                            _stprintf( szLogLine, _T("! %12s: area name %s is wrong\n"),
                                       tic.getName(),
                                       tic.getAreaName() );
							logger.rememberMode();
							logger << LOG_ERROR << szLogLine << TEXT("\n");
							logger.restoreMode();
							if (!tossingBadTics)
								MoveTicToBads( tic.getFullName(), nTicNameFrom );
                            break;
                        }
                        area = CAreasManager::getInstance()->autoCreate( tic.getAreaName(), link );

						ConstTiePtr	tie = area->getTie( link );

			            // log about new area

                        _stprintf( szLogLine, _T("+ area %s created by "), 
                                   area->getName().c_str() );
			            lstrcpy( szLogLine + lstrlen(szLogLine), tic.getFromAddress().toString().c_str() );
			            lstrcat( szLogLine, TEXT("\n") );
                        logger << szLogLine;

                        newAreasFile.Open( szNewAreasFileName, GENERIC_WRITE, OPEN_ALWAYS );
			            newAreasFile.SeekToEnd();
                        _stprintf( szLine, _T("Fileecho area %s created by "), 
							area->getName().c_str() );
			            lstrcpy( szLine + lstrlen(szLine), tic.getFromAddress().toString().c_str() );
			            lstrcat( szLine, _T("\r\n") );
			            newAreasFile.WriteString( szLine );

			            // 2. We should check for previously kept forwards
			            // 2.1. Read forwards, if have to
			            if (!m_fForwardsRead) {
				            LoadForwards();
			            }
			            int	nForwardersLinked = 0;
			            for (CForwardInfoVector::iterator iter = m_vectForwardInfo.begin(); iter != m_vectForwardInfo.end();) {
				            // 2.4.5. Compare area with just created
				            if (lstrcmpi( iter->m_szArea, tic.getAreaName() ) == 0) {
					            // 2.5. Check address
					            CAddress	addr;
								if (addr.parseString( iter->m_szLink )) {
						            // 2.5.1. Link area
									PLink	forwardLink = CLinksManager::getInstance()->getLink( addr );
						            if (forwardLink != NULL) {
										if (area->canSend( forwardLink ) || area->canReceive( forwardLink )) {
											TiePtr	newTie = new CTie();
											newTie->setLink( forwardLink );
											newTie->setRouteType( ROUTE_NONE );
											area->addTie( newTie );
											CAreasManager::getInstance()->affectAreas();

								            // 2.5.2. Count links
								            nForwardersLinked++;
								            if (nForwardersLinked) {
									            newAreasFile.WriteString( "\r\nRequestor linked:\r\n" );
								            }
								            // 2.5.3. Prepare address string
								            lstrcpy( szLine, _T("  ") );
								            lstrcat( szLine, forwardLink->getAddress().toString().c_str() );

								            // 2.5.4. Add log line

								            logger << TEXT("  linked ") << szLine+2 << TEXT("\n");

								            lstrcat( szLine, TEXT("\r\n") );
								            // 2.5.5. Add newareas.txt line
								            newAreasFile.WriteString( szLine );
							            } else {
								            // FIX: we should send message about failure
							            }
							            // remove info from forwards
							            iter = m_vectForwardInfo.erase(iter);
							            m_fForwardsChanged = true;
							            continue;
						            } else {
										logger.rememberMode();
							            logger << LOG_WARNING << TEXT("Unknown link ") << addr.toString()
								               << TEXT(" from ") << c_szForwardsFileName << TEXT("\n");
										logger.restoreMode();
						            }
                                } else {
									logger.rememberMode();
						            logger << LOG_WARNING << TEXT("Failed to parse address ") << iter->m_szLink
										   << TEXT(" from ") << c_szForwardsFileName << TEXT("\n");
									logger.restoreMode();
					            }
				            }
				            ++iter;
			            }
			            if (nForwardersLinked > 0) {
				            newAreasFile.WriteString( "\r\n\r\n" );
			            }

                        newAreasFile.Close();
                    } else {
                        _stprintf( szLogLine, TEXT("! %12s: unknown area %s\n"),
                                  tic.getName(),
                                  tic.getAreaName() );
						logger.rememberMode();
                        logger << LOG_ERROR << szLogLine;
						logger.restoreMode();
						if (!tossingBadTics)
							MoveTicToBads( tic.getFullName(), nTicNameFrom );
                        break;
                    }
                }

                // handle replaces if needed
                fReplaceDone = false;
				if (commonData->doHandleReplaces() && *(tic.m_szReplaces) != _T('\0')) {
                    lstrcpy( szDestFileName, area->getWorkingPath().c_str() );
                    lstrcat( szDestFileName, tic.m_szReplaces );
                    // check for file existance
                    if (_taccess( szDestFileName, 0 ) == 0) {
                        // remove file
                        if (_tunlink(szDestFileName) != 0) {
                            _stprintf( szReplaceLine,
                                       _T("? %12s: can't delete replaced file %s\n"),
                                       tic.getName(),
                                       szDestFileName );
                        } else {
                            area->clearDescFile( tic.m_szReplaces );
                            _stprintf( szReplaceLine,
                                       _T("                replaced file %s\n"),
                                       tic.m_szReplaces );
                        }
                        fReplaceDone = true;
                    }
                }
                // move file into fecho directory
		        lstrcpy( szDestFilePath, area->getWorkingPath().c_str() );
		        lstrcpy( szDestFileName, szDestFilePath );
                if (*tic.m_szFullName != _T('\0')) {
                    lstrcat( szDestFileName, tic.m_szFullName );
                } else {
                    lstrcat( szDestFileName, tic.getFileName() );
                }
				
				// avoid overriting of a file with long name by file with the same short name
				LongNameKeeper	longNameKeeper;
				if (longNameKeeper.keep( szDestFileName ) == LongNameKeeper::ONLY_ONE_NAME) {
					// file has only one name, not two names (short and long)
					RenameFilePlan	renamePlan = RFP_NONE;
					if (area->doesRenameFilePlanSet()) {
						renamePlan = area->getRenameFilePlan();
					} else {
						renamePlan = CCommonData::getInstance()->getRenameFilePlan();
					}
					if (renamePlan != RFP_NONE) {
						area->renameFile( szDestFileName, renamePlan );
					}
				}

                if (MyMoveFile( szFullFileName, szDestFileName ) != FO_SUCCEEDED) {
                    _stprintf( szLogLine,
                             _T("! %12s: failed to move file %s into file %s\n"),
                             tic.getName(), tic.getFileName(),
                             szDestFileName );
					logger.rememberMode();
                    logger << LOG_ERROR << szLogLine;
					logger.restoreMode();
                    break;
                }
                area->increaseTraffic( (FindFileData.nFileSizeHigh * MAXDWORD) +
                                       FindFileData.nFileSizeLow );
                fNewFilesArrived = true;
                // output information in logfile and on screen, if needed
                _stprintf( szLogLine, _T("+ %12s: %-12s in %s\n"),
                           tic.getName(), tic.getFileName(),
                           area->getName().c_str() );

                logger << szLogLine;

                if (fReplaceDone) {
                    logger << szReplaceLine;
                }

                // update descriptions
				/*

				OLD IMPLEMENTATION

                if (*tic.m_szFullName != _T('\0')) {
                    logger << TEXT("                renamed to \"") << tic.m_szFullName << _T("\"\n");
                    area->updateDescFile( tic.m_szFullName, tic.m_szDesc, fileSize, 
                                          tic.m_listLDesc );
                } else {
                    area->updateDescFile( tic.getFileName(), tic.m_szDesc, fileSize, tic.m_listLDesc );
                }
				*/
				// NEW IMPLEMENTATION (in progress):
				tic.updateDescFile( *area );

                TreatProcessedTic( tic.getFullName(), nTicNameFrom,
                                   processedTicsNamer );

				// before any forwarding ensure that Tic is STRONGLY valid
				tic.ensureValidity();

                // tic is processed, lets treat the file

                // update newfiles list

		        // Area
		        ConcatStrings( szLine, _T("Area "), area->getName().c_str() );
		        lstrcat( szLine, _T("\r\n") );
                newFilesFile.WriteString( szLine );

		        // File
		        lstrcpy( szLine, _T("File ") );
                if (*tic.m_szFullName != _T('\0')) {
                    lstrcat( szLine, tic.m_szFullName );
                } else {
                    lstrcat( szLine, tic.getFileName() );
                }
		        lstrcat( szLine, _T("\r\n") );
		        newFilesFile.WriteString( szLine );

		        // Size
		        _stprintf( szLine, _T("Size %lu\r\n"), 
			               (FindFileData.nFileSizeHigh * MAXDWORD) + FindFileData.nFileSizeLow );
		        newFilesFile.WriteString( szLine );

		        // Desc
		        _stprintf( szLine, _T("Desc %s \r\n"), tic.m_szDesc );
		        newFilesFile.WriteString( szLine );

		        // Origin
		        lstrcpy( szLine, _T("Origin ") );
		        lstrcat( szLine, tic.getOriginAddress().toStringWithoutDomain().c_str() );
		        lstrcat( szLine, _T("\r\n") );
		        newFilesFile.WriteString( szLine );

		        // Uplink
		        lstrcpy( szLine, _T("Uplink ") );
		        lstrcat( szLine, link->getAddress().toStringWithoutDomain().c_str() );
		        lstrcat( szLine, _T("\r\n") );
		        newFilesFile.WriteString( szLine );

		        // keep file name in area's vector
                if (*tic.m_szFullName != _T('\0')) {
			        area->rememberNewFile( tic.m_szFullName );
                } else {
			        area->rememberNewFile( tic.m_szFile );
                }

                // add sender to seenby list, if absent
				CTic::AddressList::iterator	iSeenby;
		        for (iSeenby = tic.getSeenByList().begin(); iSeenby != tic.getSeenByList().end(); iSeenby++) {
			        if (link->getAddress() == (*iSeenby))
				        break;
		        }
                if (iSeenby == tic.getSeenByList().end()) {
			        tic.getSeenByList().push_back( link->getAddress() );
                }
                // add us to seenby list, if absent
		        const CAddress&	ourAddr = link->getOurAddress();
				for (iSeenby = tic.getSeenByList().begin(); iSeenby != tic.getSeenByList().end(); iSeenby++) {			
			        if (ourAddr == (*iSeenby))
				        break;
		        }
				if (iSeenby == tic.getSeenByList().end()) {
			        tic.getSeenByList().push_back( ourAddr );
                }
                // add absent AKAs if AllAKASeen switched to Yes
				if (commonData->areAllAKASeen()) {
					for (index = 0; index <= commonData->getSysopAKACount(); index++) {
                        AKASeen[index] = 0;
                    }
		            for (iSeenby = tic.getSeenByList().begin(); iSeenby != tic.getSeenByList().end(); iSeenby++) {
                        // check main addr
						if (commonData->getSysopAddress() == (*iSeenby)) {
                            AKASeen[commonData->getSysopAKACount()] = 1;
                        }
                        // check AKAs
                        for (index = 0; index < commonData->getSysopAKACount(); index++) {
							if (commonData->getSysopAKA(index) == (*iSeenby)) {
                                AKASeen[index] = 1;
                                break;
                            }
                        }
		            }
                    // add absent addresses
					for (index = 0; index < commonData->getSysopAKACount(); index++) {
                        if (AKASeen[index] == 0) {
                            tic.getSeenByList().push_back( commonData->getSysopAKA(index) );
                        }
                    }
                    if (AKASeen[commonData->getSysopAKACount()] == 0) {
                        tic.getSeenByList().push_back( commonData->getSysopAddress() );
                    }
                }
                // do tics for all connected links absent in seenbys
                // first of all prepare list of links
				int	tiesCount = area->getTiesCount();
				int	i;
				for (i = 0; i < tiesCount; i++) {
					TiePtr	tie      = area->getTie( i );
					PLink	someLink = tie->getLink();
                    // skip passive or unavailable seenby
                    if (someLink->getActivity() != ACTIVITY_USUAL)
                        continue;
                    // skip sender
                    if (someLink->getAddress() == link->getAddress())
                        continue;
					if (!area->canReceive( someLink ))
						continue;
                    // check seenby
		            for (iSeenby = tic.getSeenByList().begin(); iSeenby != tic.getSeenByList().end(); iSeenby++) {
			            if (someLink->getAddress() == (*iSeenby))
				            break;
		            }
                    if (iSeenby != tic.getSeenByList().end())
                        continue;
                    // add link to list
                    // doesn't matter what is the value of RouteType
		            forwards.push_back( someLink );
                }
                // now do tics
                if (!forwards.empty())
                    fTossedForLinks = true;
				for (index = 0; index < forwards.size(); index++) {
                    tic.Forward( forwards[index], area, szDestFilePath, forwards );
                }
                // free forwards list
		        forwards.clear();

		        // create flags
		        {
                    if (*tic.m_szFullName != TEXT('\0')) {
			            lstrcpy( szLine, tic.m_szFullName );
                    } else {
			            lstrcpy( szLine, tic.m_szFile );
                    }

			        // 1) check area flags

			        int	count = area->getFileFlagCount();

			        for (i = 0; i < count; i++) {
						CMaskFileFlag&	flag = area->getFileFlag(i);
				        if (!flag.isUsed() &&
					        CheckMask( szLine, flag.getMask().c_str() ))
				        {
					        if (CreateFileFlag( flag.getFileName().c_str()) ) {
						        flag.setUsed( true );

						        logger << TEXT("file flag created: ")
							           << flag.getFileName() << TEXT("\n");
					        } else {
								logger.rememberMode();
						        logger << LOG_ERROR
									   << TEXT("failed to create file flag: ")
							           << flag.getFileName() << TEXT("\n");
								logger.restoreMode();
					        }
				        }
			        }

			        // 2) check common flags
					createFileFlagsOnFile( szLine );
		        }

                break;
            } while (true);

	        // check for warn data for sysop
	        if (!warnings.empty() || !errors.empty()) {
		        _stprintf( szLogLine, TEXT("Notification conserning %12s"),
			              tic.getName() );
		        bodyLines.push_back( TEXT("") );
		        bodyLines.push_back( szLogLine );
		        if (!warnings.empty()) {
			        bodyLines.push_back( TEXT("") );
			        bodyLines.push_back( TEXT("  Warnings:") );
			        bodyLines.push_back( TEXT("") );
			        AppendVector( bodyLines, warnings );
			        warnings.clear();
		        }
		        if (!errors.empty()) {
			        bodyLines.push_back( TEXT("") );
			        bodyLines.push_back( TEXT("  Errors:") );
			        bodyLines.push_back( TEXT("") );
			        AppendVector( bodyLines, errors );
			        errors.clear();
		        }
	        }
        } else {
            _stprintf( szLogLine, _T("! %12s: error parsering file\n"),
                     tic.getName() );
			logger.rememberMode();
            logger << LOG_ERROR << szLogLine << TEXT("                ")
				   << tic.GetLastError() << TEXT("\n");
			logger.restoreMode();
        }

		logger.flush();

	    itTic++;
    }
    newFilesFile.Close();
    // check if newfiles.txt has no data (zero size) and remove it if it hasn't
    ConcatStrings( szFullFileName, m_szStartupPath, _T("newfiles.txt") );
    hFindFile = FindFirstFile( szFullFileName, &FindFileData );
    fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
    if (fFileFound) {
        FindClose( hFindFile );
        if (((FindFileData.nFileSizeHigh * MAXDWORD) +
             FindFileData.nFileSizeLow) == 0)
            _tunlink( szFullFileName);
    }
    // save traffic, if changed
    if (fNewFilesArrived) {
	    ConcatStrings( szFullFileName, m_szStartupPath, _T("traffic.dat") );
	    CFile	trafficFile;
        if (!trafficFile.Open( szFullFileName, GENERIC_WRITE, CREATE_ALWAYS )) {
			logger.rememberMode();
			logger << LOG_ERROR << TEXT("! Error writing traffic base file ")
			       << szFullFileName << TEXT("\n");
			logger.restoreMode();
        } else {
            if (*szTimeLine == _T('\0')) {
                time_t curTime = time(NULL);
                trafficFile.WriteString( ctime( &curTime ) );
            } else {
                trafficFile.WriteString( szTimeLine );
		        trafficFile.WriteString( "\r\n" );
	        }

			int	areasCount = CAreasManager::getInstance()->getAreasCount();
			for (int i = 0; i < areasCount; i++) {
				PArea	area = CAreasManager::getInstance()->getArea( i );
			    if (area->getTraffic() > 0) {
				    _stprintf( szLine, 
					       _T("%s %lu\r\n"),
					       area->getName().c_str(),
					       area->getTraffic() );
				    trafficFile.WriteString( szLine );
			    }
	        }
            trafficFile.Close();
        }
        // check if traffic.dat has no data (zero size) and remove it if it hasn't
        hFindFile = FindFirstFile( szFullFileName, &FindFileData );
        fFileFound = (hFindFile != INVALID_HANDLE_VALUE);
        if (fFileFound) {
            FindClose( hFindFile );
            if (((FindFileData.nFileSizeHigh * MAXDWORD) +
                 FindFileData.nFileSizeLow) == 0)
                _tunlink( szFullFileName );
        }
    }
    delete [] AKASeen;

    // warn sysop if have to
    if (!bodyLines.empty()) {
		warnMessage.setFromAddress( CCommonData::getInstance()->getSysopAddress() );
	    warnMessage.setToAddress( CCommonData::getInstance()->getSysopAddress() );
		warnMessage.setSenderName( CCommonData::getInstance()->getSysopName().c_str() );
	    warnMessage.setReceiverName( "Adrenalin" );
	    warnMessage.Reply( ATTR_PRIVATE | ATTR_LOCAL, 0, "Notification", bodyLines );
    }

    // create flag, if tossed
	if (fTossedForLinks) {
		createFileFlagsOnEvent( EE_TOSS );
    }

    if (fNewFilesArrived) {
	    UpdateFilesDB();
    }
}


void CAdrenalin::TreatProcessedTic(LPCTSTR fullTicName, int onlyNameFrom, 
                                   CTicNamer& processedTicsNamer)
{
    TCHAR   logLine[LINE_SIZE];

	LoggerPtr	logger;

	const tstring& processedTicsPath = CCommonData::getInstance()->getProcessedTicsPath();

	if (processedTicsPath.empty()) {
        // no directory specified for processed tics
        // lets remove the tic
        if (_tunlink( fullTicName ) != 0) {
            wsprintf( logLine, _T("? %12s: can't delete TIC\n"),
                      fullTicName + onlyNameFrom );
            logger << LOG_ERROR << logLine;
        }
    }
    else {
        // special directory specified for processed tics
        // lets try to move the tic there

	    // form tic name
        TCHAR   newTicName[MAX_PATH];
	    wsprintf( newTicName, TEXT("%sto%06lu.tic"), processedTicsPath.c_str(),
                  processedTicsNamer.getFreeNumber() );
		if (MyMoveFile(fullTicName, newTicName) == FO_SUCCEEDED) {
            // TIC was moved
            wsprintf( logLine, _T("                after tossing TIC was moved to %s\n"),
                      newTicName + processedTicsPath.length() );
			logger << (CCommonData::getInstance()->mustShowGoOn() ? LOG_LOUD : LOG_SILENT) << logLine;
        }
        else {
            // We failed to move TIC
            wsprintf( logLine, _T("?               failed to move TIC to %s\n"),
                      newTicName + processedTicsPath.length() );
			logger << LOG_ERROR << logLine;
        }
    }
}

/*

bool CAdrenalin::processTic(CTic& tic, const tstring& filePath) {
	// check sender
	
	PLink	sender = CLinksManager::getInstance()->getLink( tic.getSenderAddress() );
	if (sender == NULL) {
		// TODO: log unknown sender
		return false;
	}

	// check sender activity

	if (sender->getActivity() == ACTIVITY_UNAVAILABLE) {
		// TODO: log inactive link
		return false;
	}

	// check password
	if (lstrcmpi( tic.getPassword(), sender->getPassword().c_str() ) != 0) {
		// TODO: log incorrect password
		return false;
	}

    // form full file name of inbound file
	tstring	fullFileName( filePath );
	fullFileName += tic.getFileName();

    // check file
	WIN32_FIND_DATA	fileData;
    HANDLE			hFile = FindFirstFile( fullFileName.c_str(), &fileData );
    if (hFile == INVALID_HANDLE_VALUE) {
		// TODO: log that tic wasn't found
		// TODO: move old tic to bad tics
		return false;
    }
    FindClose( hFile );

	// check file finished or not
	if (fileData.dwFileAttributes &
		(FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN |
		 FILE_ATTRIBUTE_OFFLINE   | FILE_ATTRIBUTE_READONLY |
		 FILE_ATTRIBUTE_TEMPORARY))
	{
		// TODO: log that file is not accessible
		return false;
	}

    // check size
    int	fileSize = ((fileData.nFileSizeHigh * MAXDWORD) + fileData.nFileSizeLow);

	if (tic.getSize() != -1) {
		if (tic.getSize() != fileSize) {
			// TODO: log that file size from tic differs from real file size
			return false;
		}
    } else {
		tic.setSize( fileSize );
    }

	// TODO: check CRC32

    // check area
	PArea	area = CAreasManager::getInstance()->getArea( tic.getAreaName() );
	TiePtr	tie  = NULL;
	if (area != NULL) {
		tie = area->getTie( sender );
        if (tie == NULL) {
			// TODO: log that sender is not linked to the area
            return false;
        }
		if (!area->canSend( sender )) {
			// TODO: log that sender cannot send into area
			return false;
		}
    } else {
		// we have no such area, let's check whether sender is allowed to create it
		if (!sender->canCreateAreas()) {
			// TODO: log unknown area
			return false;
		}
		if (CArea::isValidName(tic.getAreaName())) {
			// TODO: log that area name is invalid
            return false;
        }
		area = CAreasManager::getInstance()->autoCreate( tic.getAreaName(), sender );
		assert( area != NULL );
		tie  = area->getTie( sender );
		assert( tie != NULL );

		// TODO: log that new area created
		// TODO: write info about new area

		// TODO: check previously kept forwards
	}

	// TODO: handle replaces

	// TODO: move file into the fileecho directory

	// TODO: update descriptions

	return true;
}

*/