#include "pch.h"
#include "Logger.h"
#include "ConfigWriter.h"

#include "SuspendedSAX.h"

using namespace	miniSAX;

SuspendedHandler::SuspendedHandler(CMiniSAXParser* parser, SuspendedFiles* suspendedFiles)
    : m_parser(parser), m_state(ROOT_STATE), m_entriesHolder(), m_suspendedFiles(suspendedFiles), 
	  m_currentFile(NULL)
{
}

SuspendedHandler::~SuspendedHandler() {
	if (m_currentFile != NULL) {
		delete m_currentFile;
	}
}

static const char* SUSPENDEDFILES_LEXEM   = "suspendedFiles";
static const char* FILE_LEXEM             = "file";
static const char* NAME_LEXEM             = "name";
static const char* AREA_LEXEM             = "area";
static const char* DESCRIPTION_LEXEM      = "description";
static const char* NOCOPY_LEXEM           = "noCopy";
static const char* REPLACES_LEXEM         = "replaces";
static const char* DATETIME_LEXEM         = "dateTime";
static const char* ERROR_LEXEM            = "error";
static const char* DELETEAFTERHATCH_LEXEM = "deleteAfterHatch";


void SuspendedHandler::startElement(const char* name, CAttributeList& attributes)
    throw (CMiniSAXException)
{
    // the main idea is:
    //   we try to change state according to the element name;
    //   at the end of the method we check whether state changed or not;
    //   if not then the element is unexpected

    // remember state
    HandleState oldState = m_state;

    switch (m_state) {

    case ROOT_STATE:

        if (strcmp( name, SUSPENDEDFILES_LEXEM ) == 0) {
            checkDup( SUSPENDEDFILES_LEXEM );
            m_state = SUSPENDEDFILES_STATE;
            break;
        }
        break;

    case SUSPENDEDFILES_STATE:

        if (strcmp( name, FILE_LEXEM ) == 0) {
            rememberEntry( FILE_LEXEM );

			m_currentFile = new SuspendedFile();

            m_state = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case SUSPENDEDFILES_FILE_STATE:

        if (strcmp( name, NAME_LEXEM ) == 0) {
            checkDup( NAME_LEXEM );
            m_state = FILE_NAME_STATE;
            break;
        }
        if (strcmp( name, DATETIME_LEXEM ) == 0) {
            checkDup( DATETIME_LEXEM );

			int	year, month, day, hour, minute, second;

            const CAttribute* yearAttr = attributes.getAttribute( "year" );
            if (yearAttr == NULL) {
				m_parser->raiseEx( "year attribute absent" );
			} else {
				parseUnsignedInt( yearAttr->getValue(), year );
			}

            const CAttribute* monthAttr = attributes.getAttribute( "month" );
            if (monthAttr == NULL) {
				m_parser->raiseEx( "month attribute absent" );
			} else {
                parseUnsignedInt( monthAttr->getValue(), month );
				if (month < 1 || month > 12) {
					m_parser->raiseEx( "Invalid month value" );
				}
            }
            const CAttribute* dayAttr = attributes.getAttribute( "day" );
            if (dayAttr == NULL) {
				m_parser->raiseEx( "day attribute absent" );
			} else {
                parseUnsignedInt( dayAttr->getValue(), day );
				if (day < 1 || day > 31) {
					m_parser->raiseEx( "Invalid day value" );
				}
            }
            const CAttribute* hourAttr = attributes.getAttribute( "hour" );
            if (hourAttr == NULL) {
				m_parser->raiseEx( "hour attribute absent" );
			} else {
                parseUnsignedInt( hourAttr->getValue(), hour );
				if (hour > 23) {
					m_parser->raiseEx( "Invalid hour value" );
				}
            }
            const CAttribute* minuteAttr = attributes.getAttribute( "minute" );
            if (minuteAttr == NULL) {
				m_parser->raiseEx( "minute attribute absent" );
			} else {
                parseUnsignedInt( minuteAttr->getValue(), minute );
				if (minute > 59) {
					m_parser->raiseEx( "Invalid minute value" );
				}
            }
            const CAttribute* secondAttr = attributes.getAttribute( "second" );
            if (secondAttr == NULL) {
				m_parser->raiseEx( "second attribute absent" );
			} else {
                parseUnsignedInt( secondAttr->getValue(), second );
				if (second > 59) {
					m_parser->raiseEx( "Invalid second value" );
				}
            }

			SYSTEMTIME	dateTime;
			::ZeroMemory( &dateTime, sizeof(dateTime) );
			dateTime.wYear   = year;
			dateTime.wMonth  = month;
			dateTime.wDay    = day;
			dateTime.wHour   = hour;
			dateTime.wMinute = minute;
			dateTime.wSecond = second;

			m_currentFile->setReleaseDateTime( dateTime );

            m_state = FILE_DATETIME_STATE;
            break;
        }
        if (strcmp( name, AREA_LEXEM ) == 0) {
            checkDup( AREA_LEXEM );
            m_state = FILE_AREA_STATE;
            break;
        }
        if (strcmp( name, DESCRIPTION_LEXEM ) == 0) {
            checkDup( DESCRIPTION_LEXEM );
            m_state = FILE_DESCRIPTION_STATE;
            break;
        }
        if (strcmp( name, NOCOPY_LEXEM ) == 0) {
            checkDup( NOCOPY_LEXEM );
			m_currentFile->setNoCopy( true );
            m_state = FILE_NOCOPY_STATE;
            break;
        }
        if (strcmp( name, REPLACES_LEXEM ) == 0) {
            checkDup( REPLACES_LEXEM );
            m_state = FILE_REPLACES_STATE;
            break;
        }
        if (strcmp( name, ERROR_LEXEM ) == 0) {
            checkDup( ERROR_LEXEM );
            m_state = FILE_ERROR_STATE;
            break;
        }
        if (strcmp( name, DELETEAFTERHATCH_LEXEM ) == 0) {
            checkDup( DELETEAFTERHATCH_LEXEM );
			m_currentFile->setDeleteAfterHatch( true );
            m_state = FILE_DELETEAFTERHATCH_STATE;
            break;
        }
        break;
    }
    if (oldState == m_state) {
        // we didn't change the state => we have unexpected element
        m_parser->raiseEx( "Unexpected element" );
    }
    m_parser->checkUnexpected( attributes );
}


void SuspendedHandler::endElement(const char* name)
    throw (CMiniSAXException)
{
    HandleState newState = m_state;
    switch (m_state) {

    case SUSPENDEDFILES_STATE:

        if (strcmp( name, SUSPENDEDFILES_LEXEM ) == 0) {
            newState = ROOT_STATE;
            break;
        }
        break;

    case SUSPENDEDFILES_FILE_STATE:

        if (strcmp( name, FILE_LEXEM ) == 0) {
            if (!hasEntry( NAME_LEXEM )) {
                m_parser->raiseEx( "name undefined" );
            }
            if (!hasEntry( DATETIME_LEXEM )) {
                m_parser->raiseEx( "dateTime undefined" );
            }
            if (!hasEntry( AREA_LEXEM )) {
                m_parser->raiseEx( "area undefined" );
            }
            if (!hasEntry( DESCRIPTION_LEXEM )) {
                m_parser->raiseEx( "description undefined" );
            }

			m_suspendedFiles->push_back( m_currentFile );
			m_currentFile = NULL;

            newState = SUSPENDEDFILES_STATE;
            break;
        }
        break;

    case FILE_NAME_STATE:

        if (strcmp( name, NAME_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_DATETIME_STATE:

        if (strcmp( name, DATETIME_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_AREA_STATE:

        if (strcmp( name, AREA_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_DESCRIPTION_STATE:

        if (strcmp( name, DESCRIPTION_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_NOCOPY_STATE:

        if (strcmp( name, NOCOPY_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_REPLACES_STATE:

        if (strcmp( name, REPLACES_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_ERROR_STATE:

        if (strcmp( name, ERROR_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    case FILE_DELETEAFTERHATCH_STATE:

        if (strcmp( name, DELETEAFTERHATCH_LEXEM ) == 0) {
            newState = SUSPENDEDFILES_FILE_STATE;
            break;
        }
        break;

    }
    if (newState == m_state) {
        // state not changed => we have unexpected element
        m_parser->raiseEx( "Unexpected end of element" );
    } else {
        removeEntries();
        m_state = newState;
    }
}


void SuspendedHandler::handleCharacters(const char* buffer, size_t len)
    throw (CMiniSAXException)
{
    switch (m_state) {

    case FILE_NAME_STATE:
		{
			m_currentFile->setFileName( tstring( buffer, len ) );
		}
        return;

    case FILE_AREA_STATE:
		{
			m_currentFile->setAreaName( tstring( buffer, len ) );
		}
        return;

    case FILE_DESCRIPTION_STATE:
		{
			m_currentFile->setDescription( tstring( buffer, len ) );
		}
        return;

    case FILE_REPLACES_STATE:
		{
			m_currentFile->setReplaces( tstring( buffer, len ) );
		}
        return;

    case FILE_ERROR_STATE:
		{
			m_currentFile->setError( tstring( buffer, len ) );
		}
        return;

    }
    if (!hasOnlySpaces( buffer, len )) {
        m_parser->raiseEx( "Unexpected symbols" );
    }
}


void SuspendedHandler::checkDup(LPCTSTR name, bool remember /* = true */) throw(CMiniSAXException) {
    if (hasEntry( name )) {
        m_parser->raiseEx( (std::string("Element '") + std::string(name) + std::string("' was already defined")).c_str() );
    } else {
        if (remember) {
            rememberEntry( name );
        }
    }
}


void SuspendedHandler::rememberEntry(LPCTSTR name) {
    EntryDescriptor desc;
    desc.m_name  = name;
    desc.m_state = m_state;
    m_entriesHolder.push_back( desc );
}


void SuspendedHandler::removeEntry(LPCTSTR name) {
    EntriesHolder::iterator it = getEntry(name);
    if (it != m_entriesHolder.end()) {
        m_entriesHolder.erase( it );
    }
}


SuspendedHandler::EntriesHolder::iterator SuspendedHandler::getEntry(LPCTSTR name) {
    for (EntriesHolder::iterator it = m_entriesHolder.begin();
         it != m_entriesHolder.end(); ++it)
    {
        if ((*it).m_name == name && (*it).m_state == m_state) {
            return it;
        }
    }
    return m_entriesHolder.end();
}


bool SuspendedHandler::hasEntry(LPCTSTR name) {
    return getEntry( name ) != m_entriesHolder.end();
}


void SuspendedHandler::removeEntries() {
    EntriesHolder::iterator it = m_entriesHolder.begin();
    while (it != m_entriesHolder.end()) {
        if ((*it).m_state == m_state) {
            it = m_entriesHolder.erase( it );
        } else {
            ++it;
        }
    }
}


bool SuspendedHandler::hasOnlySpaces(const char* s, size_t len) {
    while (len > 0) {
        --len;
        switch (s[len]) {
            case ' ':
            case '\n':
            case '\r':
            case '\t':
                break;
            default:
                return false;
        }
    }

    return true;
}


void SuspendedHandler::parseUnsignedInt(const char* s, int& value)
	throw(CMiniSAXException)
{
	int	scanResult = sscanf( s, "%d", &value );
	if (scanResult == 0 || scanResult == EOF || value < 0) {
		m_parser->raiseEx( "Invalid unsigned integer" );
	}
}

// all XML errors are handled inside
bool SuspendedHandler::readSuspendedFiles(const tstring& fileName, SuspendedFiles* suspendedFiles, tstring& documentEncoding) {
	bool	okey = true;

	try {
		auto_ptr<CMiniSAXParser>	parserPtr( CMiniSAXParser::create() );
		parserPtr->setHandlerEncoding( ENCODING_CP866 );
		SuspendedHandler	suspendedHandler( parserPtr.get(), suspendedFiles );

		CDataSource		dataSource( fileName.c_str() );

		parserPtr->setContentHandler( &suspendedHandler );

		parserPtr->parse( &dataSource );

		// remember document encoding if it was specified

		const char* docEncoding = parserPtr->getDocumentEncoding();
		if (docEncoding != NULL) {
			documentEncoding = docEncoding;
		}

	} catch (miniSAX::CIOException& ioEx) {
		LoggerPtr	logger;
		logger << LOG_ERROR << TEXT("Failed to read ") << fileName << TEXT(": ") 
			   << ioEx.getMessage() << TEXT("\n");
		okey = false;
	} catch (CMiniSAXException& saxEx) {
		// prepare exception message
		tstring	msg("Failed to read ");
		msg += fileName;
		msg += TEXT(": ");
		if (saxEx.getColumn() >= 0 && saxEx.getRow() >= 0) {
			char	buffer[256];
			sprintf( buffer, " [col:%d,row:%d] ", saxEx.getColumn(), saxEx.getRow() );
			msg += buffer;
		} else {
			msg += " [no position info] ";
		}
		msg += saxEx.getMessage();
		LoggerPtr	logger;
		logger << LOG_ERROR << msg << TEXT("\n");
		okey = false;
	}

	if (!okey) {
		// delete all files we've already read
		while (!suspendedFiles->empty()) {
			SuspendedFilePtr	file = suspendedFiles->at( suspendedFiles->size() - 1 );
			suspendedFiles->pop_back();
			delete file;
		}
	}

	return okey;
}

void SuspendedHandler::writeSuspendedFiles(const tstring& fileName, SuspendedFiles* suspendedFiles, const tstring& documentEncoding) {
	const char*	destEncoding = (documentEncoding.empty() ? NULL : documentEncoding.c_str());

	CConfigWriter	out( fileName, ENCODING_CP866, destEncoding );

	// prepare manipulator for entities
	CConfigWriter::CEntityResolver	er;

	try {
		out << "<suspendedFiles>\n";
		size_t	count = suspendedFiles->size();
		for (size_t	i = 0; i < count; i++) {
			SuspendedFilePtr	file = suspendedFiles->at(i);
			out << "    <file>\n"
				<< "        <name>" << er << file->getFileName() << "</name>\n"
				<< "        <area>" << er << file->getAreaName() << "</area>\n"
				<< "        <description>" << er << file->getDescription() << "</description>\n";
			if (file->isNoCopy()) {
				out << "        <noCopy />\n";
			}
			if (file->doDeleteAfterHatch()) {
				out << "        <deleteAfterHatch />\n";
			}
			if (!file->getReplaces().empty()) {
				out << "        <replaces>" << er << file->getReplaces() << "</replaces>\n";
			}
			const SYSTEMTIME&	dateTime = file->getReleaseDateTime();
			out << "        <dateTime "
				<< "year=\"" << (int)dateTime.wYear << "\" "
				<< "month=\"" << (int)dateTime.wMonth << "\" "
				<< "day=\"" << (int)dateTime.wDay << "\" "
				<< "hour=\"" << (int)dateTime.wHour << "\" "
				<< "minute=\"" << (int)dateTime.wMinute <<  "\" "
				<< "second=\"" << (int)dateTime.wSecond <<  "\" />\n";
			if (file->hasError()) {
				out << "        <error>" << er << file->getError() << "</error>\n";
			}
			out	<< "    </file>\n";
		}
		out << "</suspendedFiles>\n";
	} catch (...) {
		out.cancel();
		LoggerPtr	logger;
		logger << LOG_ERROR << TEXT("Failed to write ") << fileName << TEXT("\n");
	}

	out.close();
}

bool SuspendedHandler::doesFileExist(const tstring& fileName) {
	// check file existance and read permission
	return (_taccess( fileName.c_str(), 0 ) == 0);
}
